/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced;

import java.io.Serializable;

import org.ow2.util.pool.api.IPoolConfiguration;

/**
 * Implementation of the Pool Configuration interface.
 * @author Florent BENOIT
 */
public class PoolConfiguration implements IPoolConfiguration, Serializable {

    /**
     * Id for serializable class.
     */
    private static final long serialVersionUID = 3409409761151246408L;

    /**
     * Minimum size of the pool.
     */
    private int min = DEFAULT_MIN;

    /**
     * Maximum size of the pool.
     */
    private int max = DEFAULT_MAX;

    /**
     * Timeout for waiters that are waiting for an empty element in the pool.
     */
    private long timeout = DEFAULT_TIMEOUT;

    /**
     * Number of waiters that can wait for the given timeout.
     */
    private int maxWaiters = DEFAULT_MAX_WAITERS;

    /**
     * Spare size of the pool.
     */
    private int spare = DEFAULT_SPARE;

    /**
     * @return Maximum size of the pool.
     */
    public int getMax() {
        return max;
    }

    /**
     * Sets the Maximum size of the pool.
     * @param max the given maximum size of the pool
     */
    public void setMax(final int max) {
        this.max = max;
    }

    /**
     * @return Minimum size of the pool.
     */
    public int getMin() {
        return min;
    }

    /**
     * Sets the Minimum size of the pool.
     * @param min the given minimum size of the pool
     */
    public void setMin(final int min) {
        this.min = min;
    }

    /**
     * @return spare size of the pool.
     */
    public int getSpare() {
        return spare;
    }

    /**
     * Sets the spare size of the pool.
     * @param spare the given spare size of the pool
     */
    public void setSpare(final int spare) {
        this.spare = spare;
    }

    /**
     * @return the timeout for waiters that are waiting for an empty element in the pool.
     */
    public long getTimeout() {
        return timeout;
    }

    /**
     * Sets the timeout for waiters that are waiting for an empty element in the pool..
     * @param timeout the timeout for waiters that are waiting for an empty element in the pool.
     */
    public void setTimeout(final long timeout) {
        this.timeout = timeout;
    }

    /**
     * @return the number of waiters that can wait for the given timeout. If this number is
     * reached, there is no wait and an exception is thrown.
     */
    public int getMaxWaiters() {
        return maxWaiters;
    }

    /**
     * Sets the number of waiters that can wait for the given timeout.
     * @param maxWaiters the number of waiters that can wait for the given timeout.
     */
    public void setMaxWaiters(final int maxWaiters) {
        this.maxWaiters = maxWaiters;
    }

    /**
     * Indicates whether some other object is "equal to" this one.
     * @param  object the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     */
    @Override
    public boolean equals(final Object object) {
        if (object == null || !(object instanceof PoolConfiguration)) {
            return false;
        }
        PoolConfiguration poolConfiguration = (PoolConfiguration) object;
        return max == poolConfiguration.max
            && maxWaiters == poolConfiguration.maxWaiters
            && timeout == poolConfiguration.timeout
            && min == poolConfiguration.min
            && spare == poolConfiguration.spare;
    }

    /**
     * Returns a hash code value for the pool configuration.
     * @return  a hash code value for this pool configuration.
     */
    @Override
    public int hashCode() {
        return (int) (max + maxWaiters + timeout + min + spare);
    }

}
