/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.keepbusy;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.locks.Lock;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.NotABusyPoolItemException;
import org.ow2.util.pool.impl.enhanced.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.WaiterInterruptedException;
import org.ow2.util.pool.impl.enhanced.api.keepbusy.ShareMethod;
import org.ow2.util.pool.impl.enhanced.impl.PoolError;
import org.ow2.util.pool.impl.enhanced.impl.util.LockFactory;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.NoWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.WaitAuthorization;

/**
 * @author Gael Lalire
 * @param <E> pool item type
 */
public class KeepBusyPool<E> implements IPool<E> {

    private static final Log LOG = LogFactory.getLog(KeepBusyPool.class);

    /**
     * We keep number of user.
     */
    private Map<IIdentityReference<E>, PoolItemInfo> busyPoolItemList;

    private ReferenceQueue<E> referenceQueue;

    private ShareMethod shareMethod;

    private Lock lock;

    private IPool<E> pool;

    private ReferenceType referenceType;

    private WaitAuthorization currentWaitAuthorization;

    public KeepBusyPool(final IPool<E> pool) {
        this(pool, ReferenceType.WEAK);
    }

    public KeepBusyPool(final IPool<E> pool, final ReferenceType referenceType) {
        if (pool == null) {
            throw new IllegalArgumentException();
        }
        lock = LockFactory.createLock();
        this.referenceType = referenceType;
        this.pool = pool;
        busyPoolItemList = new HashMap<IIdentityReference<E>, PoolItemInfo>();
        shareMethod = ShareMethod.NEVER_SHARE;
        switch (referenceType) {
        case STRONG:
            break;
        case WEAK:
        default:
            referenceQueue = new ReferenceQueue<E>();
        }
        currentWaitAuthorization = new WaitAuthorization();
    }

    protected void expunge() {
        switch (referenceType) {
        case STRONG:
            return;
        case WEAK:
        default:
        }
        Reference<? extends E> reference = referenceQueue.poll();
        while (reference != null) {
            PoolItemInfo poolItemInfo = busyPoolItemList.remove(reference);
            int busyCount = poolItemInfo.getBusyCount();
            if (busyCount != 0) {
                if (LOG.isWarnEnabled()) {
                    CharArrayWriter charArrayWriter = new CharArrayWriter();
                    PrintWriter printWriter = new PrintWriter(charArrayWriter);
                    for (Throwable throwable : poolItemInfo.getThrowableList()) {
                        throwable.printStackTrace(printWriter);
                        printWriter.append("\n");
                    }
                    if (busyCount == 1) {
                        LOG.warn("A poolItem was GC while user didn''t send back their instance to pool.\n"
                                + "StackTrace :\n{0}", charArrayWriter);
                    } else {
                        LOG.warn("A poolItem was GC while {0} users didn''t send back their instance to pool.\n"
                                + "StackTraces :\n{1}", busyCount, charArrayWriter);
                    }
                }
            }
            reference = referenceQueue.poll();
        }
    }

    private E lookForAvailableInstance() throws PoolException, InterruptedException {
        try {
            E poolItem = pool.get(NoWaitControl.INSTANCE); // no wait with this call
            // if we add a new pool item then we have to
            // notify other waiters
            if (shareMethod != ShareMethod.NEVER_SHARE) {
                currentWaitAuthorization.forbidWait();
                currentWaitAuthorization = new WaitAuthorization();
                pool.signalAllWaiters();
            }
            LOG.debug("pool returns NEW {0}", poolItem);
            return poolItem;
        } catch (TimeoutPoolException e) {
            return null;
        }
    }

    // FIXME: random ? less shared ? delegate choice ?
    private E lookForSharedInstance() {
        Iterator<IIdentityReference<E>> iterator = busyPoolItemList.keySet().iterator();
        E poolItem = null;
        while (poolItem == null && iterator.hasNext()) {
            poolItem = iterator.next().getReference();
        }
        LOG.debug("pool returns SHARED {0}", poolItem);
        return poolItem;
    }

    private E lookForInstance() throws PoolException, InterruptedException {
        E poolItem;
        switch (shareMethod) {
        default:
        case NEVER_SHARE:
            poolItem = lookForAvailableInstance();
            break;
        case SHARE_AS_LAST_SOLUTION:
            poolItem = lookForAvailableInstance();
            if (poolItem == null) {
                poolItem = lookForSharedInstance();
            }
            break;
        case SHARE_IF_POSSIBLE:
            poolItem = lookForSharedInstance();
            if (poolItem == null) {
                poolItem = lookForAvailableInstance();
            }
            break;
        }
        return poolItem;
    }

    public E get(final IWaitControl timeout) throws PoolException, InterruptedException {
        E poolItem = null;
        // take sync
        lock.lock();
        try {
            expunge();
            poolItem = lookForInstance();
            if (poolItem == null) {
                boolean stopWaiting = false;
                if (!timeout.canContinueToWait()) {
                    throw new TimeoutPoolException();
                }
                while (!stopWaiting) {
                    try {
                        timeout.pushWaitAuthorization(currentWaitAuthorization);
                        // release sync
                        lock.unlock();
                        try {
                            poolItem = pool.get(timeout);
                        } finally {
                            // retake sync
                            lock.lock();
                            timeout.popWaitAuthorization();
                        }
                        if (shareMethod != ShareMethod.NEVER_SHARE) {
                            // other waiter may use the poolItem
                            // NB : the poolItem will be put in busyPoolItemList
                            // before lock released
                            currentWaitAuthorization.forbidWait();
                            currentWaitAuthorization = new WaitAuthorization();
                            pool.signalAllWaiters();
                        }
                    } catch (WaiterInterruptedException e) {
                        timeout.verifyInterrupted();
                        if (shareMethod != ShareMethod.NEVER_SHARE) {
                            poolItem = lookForSharedInstance();
                        }
                    } catch (TimeoutPoolException e) {
                        // we could suppress next line
                        if (shareMethod != ShareMethod.NEVER_SHARE) {
                            poolItem = lookForSharedInstance();
                        }
                    }
                    if (poolItem != null) {
                        stopWaiting = true;
                    } else {
                        if (!timeout.canContinueToWait()) {
                            throw new TimeoutPoolException();
                        }
                    }
                }
            }
            // here pool item is not null
            IdentityReference<E> strongReference = new IdentityReference<E>(poolItem);
            PoolItemInfo poolItemInfo = busyPoolItemList.get(strongReference);
            if (poolItemInfo == null) {
                poolItemInfo = new PoolItemInfo();
                IIdentityReference<E> toPutIdentityReference;
                switch (getReferenceType()) {
                case STRONG:
                    toPutIdentityReference = strongReference;
                    break;
                case WEAK:
                default:
                    toPutIdentityReference = new IdentityWeakReference<E>(poolItem, referenceQueue);
                }
                busyPoolItemList.put(toPutIdentityReference, poolItemInfo);
            } else {
                poolItemInfo.setBusyCount(poolItemInfo.getBusyCount() + 1);
            }
            poolItemInfo.addStackTraceElements(new PoolStackTrace());
        } finally {
            // release sync
            lock.unlock();
        }
        // here we have no lock
        return poolItem;
    }

    /**
     * @param poolItem
     * @throws PoolException
     */
    public void put(final E poolItem) throws NotABusyPoolItemException {
        if (poolItem == null) {
            throw new IllegalArgumentException();
        }

        lock.lock();
        try {
            expunge();
            PoolItemInfo poolItemInfo = busyPoolItemList.get(new IdentityReference<E>(poolItem));
            if (poolItemInfo == null) {
                throw new NotABusyPoolItemException();
            }
            int sharedCountValue = poolItemInfo.getBusyCount();
            if (sharedCountValue == 1) {
                busyPoolItemList.remove(new IdentityReference<E>(poolItem));
                if (!poolItemInfo.isRemoved()) {
                    lock.unlock();
                    try {
                        pool.put(poolItem);
                    } finally {
                        lock.lock();
                    }
                }
                poolItemInfo.setBusyCount(0);
            } else {
                poolItemInfo.setBusyCount(sharedCountValue - 1);
            }
            if (shareMethod != ShareMethod.NEVER_SHARE) {
                currentWaitAuthorization.forbidWait();
                currentWaitAuthorization = new WaitAuthorization();
                pool.signalAllWaiters();
            }
        } finally {
            lock.unlock();
        }

    }

    protected ShareMethod getShareMethod0() {
        return shareMethod;
    }

    /**
     * @return the shareMethod
     */
    public ShareMethod getShareMethod() {
        lock.lock();
        try {
            return shareMethod;
        } finally {
            lock.unlock();
        }
    }

    /**
     * @param shareMethod the shareMethod to set
     */
    public void setShareMethod(final ShareMethod shareMethod) {
        lock.lock();
        try {
            this.shareMethod = shareMethod;
        } finally {
            lock.unlock();
        }
    }

    public void signalAllWaiters() {
        lock.lock();
        try {
            currentWaitAuthorization.forbidWait();
            currentWaitAuthorization = new WaitAuthorization();
            pool.signalAllWaiters();
        } finally {
            lock.unlock();
        }
    }

    public void remove(final E poolItem) throws NotABusyPoolItemException {
        if (poolItem == null) {
            throw new IllegalArgumentException();
        }
        lock.lock();

        try {
            expunge();
            PoolItemInfo poolItemInfo = busyPoolItemList.get(new IdentityReference<E>(poolItem));
            if (poolItemInfo == null) {
                throw new NotABusyPoolItemException();
            }
            if (!poolItemInfo.isRemoved()) {
                lock.unlock();
                try {
                    pool.remove(poolItem);
                } finally {
                    lock.lock();
                }
                poolItemInfo.setRemoved(true);
            }
            int sharedCountValue = poolItemInfo.getBusyCount();
            if (sharedCountValue == 1) {
                busyPoolItemList.remove(new IdentityReference<E>(poolItem));
                poolItemInfo.setBusyCount(0);
            } else {
                poolItemInfo.setBusyCount(sharedCountValue - 1);
            }
        } finally {
            lock.unlock();
        }
    }

    /**
     * @return the busyPoolItemList
     */
    protected Map<IIdentityReference<E>, PoolItemInfo> getBusyPoolItemList() {
        return busyPoolItemList;
    }

    /**
     * @return the lock
     */
    protected final Lock getLock() {
        return lock;
    }

    public void removeAll() {
        lock.lock();
        try {
            expunge();
            for (Entry<IIdentityReference<E>, PoolItemInfo> entry : busyPoolItemList.entrySet()) {
                PoolItemInfo poolItemInfo = entry.getValue();
                E poolItem = entry.getKey().getReference();
                if (poolItem != null && !poolItemInfo.isRemoved()) {
                    lock.unlock();
                    try {
                        pool.remove(poolItem);
                    } catch (NotABusyPoolItemException e) {
                        throw new PoolError();
                    } finally {
                        lock.lock();
                    }
                    poolItemInfo.setRemoved(true);
                }
            }
        } finally {
            lock.unlock();
        }
    }

    protected ReferenceQueue<E> getReferenceQueue() {
        return referenceQueue;
    }

    protected ReferenceType getReferenceType() {
        return referenceType;
    }

    protected WaitAuthorization getCurrentWaitAuthorization() {
        return currentWaitAuthorization;
    }

    protected void setCurrentWaitAuthorization(final WaitAuthorization currentWaitAuthorization) {
        this.currentWaitAuthorization = currentWaitAuthorization;
    }

}
