/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.keepbusy.clue;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.locks.Lock;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.WaiterInterruptedException;
import org.ow2.util.pool.impl.enhanced.api.clue.IClueManager;
import org.ow2.util.pool.impl.enhanced.api.clue.ICluePool;
import org.ow2.util.pool.impl.enhanced.api.keepbusy.ShareMethod;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.IIdentityReference;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.IdentityReference;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.IdentityWeakReference;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.KeepBusyPool;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.PoolItemInfo;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.PoolStackTrace;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.NoWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.WaitAuthorization;


/**
 * @author Gael Lalire
 * @param <E> pool item type
 * @param <C> clue type type of clue
 */
public class KeepBusyCluePool<E, C> extends KeepBusyPool<E> implements ICluePool<E, C> {

    private static final Log LOG = LogFactory.getLog(KeepBusyCluePool.class);

    /**
     * We keep number of user.
     */
    private Map<IIdentityReference<E>, PoolItemInfo> busyPoolItemList;

    private IClueManager<E, C> clueManager;

    private ICluePool<E, C> cluePool;

    private Lock lock;

    public KeepBusyCluePool(final ICluePool<E, C> cluePool, final IClueManager<E, C> clueManager) {
        super(cluePool);
        if (clueManager == null) {
            throw new IllegalArgumentException();
        }
        busyPoolItemList = getBusyPoolItemList();
        this.cluePool = cluePool;
        this.clueManager = clueManager;
        lock = getLock();
    }

    /**
     * Because not same behavior.
     */
    @Override
    public E get(final IWaitControl timeout) throws PoolException, InterruptedException {
        return get(null, timeout);
    }

    private E lookForAvailableInstance(final C clue) throws PoolException, InterruptedException {
        try {
            E poolItem = cluePool.get(clue, NoWaitControl.INSTANCE); // no wait with this call
            // if we add a new pool item then we have to
            // notify other waiters
            if (getShareMethod0() != ShareMethod.NEVER_SHARE) {
                cluePool.signalAllWaiters();
            }
            LOG.debug("clue pool returns NEW {0}", poolItem);
            return poolItem;
        } catch (TimeoutPoolException e) {
            return null;
        }
    }

    private E lookForSharedInstance(final C clue) {
        Iterator<IIdentityReference<E>> iterator = busyPoolItemList.keySet().iterator();
        while (iterator.hasNext()) {
            E poolItem = iterator.next().getReference();
            if (poolItem == null) {
                continue;
            }
            if (clueManager.tryMatch(poolItem, clue)) {
                LOG.debug("clue pool returns SHARED {0}", poolItem);
                return poolItem;
            }
        }
        return null;
    }

    private E lookForInstance(final C clue) throws PoolException, InterruptedException {
        E poolItem;
        switch (getShareMethod0()) {
        default:
        case NEVER_SHARE:
            poolItem = lookForAvailableInstance(clue);
            break;
        case SHARE_AS_LAST_SOLUTION:
            poolItem = lookForAvailableInstance(clue);
            if (poolItem == null) {
                poolItem = lookForSharedInstance(clue);
            }
            break;
        case SHARE_IF_POSSIBLE:
            poolItem = lookForSharedInstance(clue);
            if (poolItem == null) {
                poolItem = lookForAvailableInstance(clue);
            }
            break;
        }
        return poolItem;
    }

    public E get(final C clue, final IWaitControl timeout) throws PoolException, InterruptedException {
        E poolItem = null;
        // take sync
        lock.lock();
        try {
            expunge();
            poolItem = lookForInstance(clue);
            if (poolItem == null) {
                boolean stopWaiting = false;
                if (!timeout.canContinueToWait()) {
                    throw new TimeoutPoolException();
                }
                while (!stopWaiting) {
                    try {
                        timeout.pushWaitAuthorization(getCurrentWaitAuthorization());
                        // release sync
                        lock.unlock();
                        try {
                            poolItem = cluePool.get(clue, timeout);
                        } finally {
                            // retake sync
                            lock.lock();
                            timeout.popWaitAuthorization();
                        }
                        if (getShareMethod0() != ShareMethod.NEVER_SHARE) {
                            // other waiter may use the poolItem
                            // NB : the poolItem will be put in busyPoolItemList before lock released
                            getCurrentWaitAuthorization().forbidWait();
                            setCurrentWaitAuthorization(new WaitAuthorization());
                            cluePool.signalAllWaiters();
                        }
                    } catch (WaiterInterruptedException e) {
                        timeout.verifyInterrupted();
                        if (getShareMethod0() != ShareMethod.NEVER_SHARE) {
                            poolItem = lookForSharedInstance(clue);
                        }
                    } catch (TimeoutPoolException e) {
                        // we could suppress next line
                        if (getShareMethod0() != ShareMethod.NEVER_SHARE) {
                            poolItem = lookForSharedInstance(clue);
                        }
                    }
                    if (poolItem != null) {
                        stopWaiting = true;
                    } else {
                        if (!timeout.canContinueToWait()) {
                            throw new TimeoutPoolException();
                        }
                    }
                }
            }
            // here pool item is not null
            IdentityReference<E> strongReference = new IdentityReference<E>(poolItem);
            PoolItemInfo poolItemInfo = busyPoolItemList.get(strongReference);
            if (poolItemInfo == null) {
                poolItemInfo = new PoolItemInfo();
                IIdentityReference<E> toPutIdentityReference;
                switch (getReferenceType()) {
                case STRONG:
                    toPutIdentityReference = strongReference;
                    break;
                case WEAK:
                default:
                    toPutIdentityReference = new IdentityWeakReference<E>(poolItem, getReferenceQueue());
                }
                busyPoolItemList.put(toPutIdentityReference, poolItemInfo);
            } else {
                poolItemInfo.setBusyCount(poolItemInfo.getBusyCount() + 1);
            }
            poolItemInfo.addStackTraceElements(new PoolStackTrace());
        } finally {
            // release sync
            lock.unlock();
        }
        // here we have no lock
        return poolItem;
    }

}
