/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.limited;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.NotABusyPoolItemException;
import org.ow2.util.pool.impl.enhanced.api.PoolException;


/**
 *
 * @author Gael Lalire
 * @param <E> pool item type
 */
public class LimitedWaiterPool<E> implements IPool<E> {

    private static final Log LOG = LogFactory.getLog(LimitedWaiterPool.class);

    public static final int NO_LIMIT_WAITER = -1;

    private Object mutex;

    private int maxWaiter;

    private int currentWaiter;

    private IPool<E> pool;

    public LimitedWaiterPool(final IPool<E> pool) {
        this(pool, NO_LIMIT_WAITER);
    }

    public LimitedWaiterPool(final IPool<E> pool, final int maxWaiter) {
        if (maxWaiter < 0 && maxWaiter != NO_LIMIT_WAITER) {
            throw new IllegalArgumentException();
        }
        mutex = new Object();
        this.pool = pool;
        this.maxWaiter = maxWaiter;
    }

    public E get(final IWaitControl timeout) throws PoolException, InterruptedException {
        addWaiter();
        try {
            return pool.get(timeout);
        } finally {
            removeWaiter();
        }
    }

    protected void addWaiter() throws TooManyWaiterException {
        synchronized (mutex) {
            if (maxWaiter != NO_LIMIT_WAITER && currentWaiter >= maxWaiter) {
                throw new TooManyWaiterException();
            }
            currentWaiter++;
            LOG.debug("there is {0} waiters ", currentWaiter);
        }
    }

    protected void removeWaiter() throws TooManyWaiterException {
        synchronized (mutex) {
            currentWaiter--;
            if (currentWaiter == 0) {
                mutex.notifyAll();
            }
            LOG.debug("there is {0} waiters ", currentWaiter);
        }
    }

    public void put(final E e) throws NotABusyPoolItemException {
        pool.put(e);
    }

    /**
     * @param maxWaiter the maxWaiter to set
     */
    public void setMaxWaiter(final int maxWaiter) {
        if (maxWaiter < 0) {
            throw new IllegalArgumentException();
        }
        synchronized (mutex) {
            this.maxWaiter = maxWaiter;
        }
    }

    public void signalAllWaiters() {
        pool.signalAllWaiters();
    }

    public void remove(final E poolItem) throws NotABusyPoolItemException {
        pool.remove(poolItem);
    }

    public void waitEmptyWaiterInPool() throws InterruptedException {
        synchronized (mutex) {
            while (currentWaiter != 0) {
                mutex.wait();
            }
        }
    }

    /**
     * @return maxWaiter
     */
    public int getMaxWaiter() {
        synchronized (mutex) {
            return this.maxWaiter;
        }
    }

}
