/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.listener;

import java.util.ArrayList;
import java.util.List;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.NotABusyPoolItemException;
import org.ow2.util.pool.impl.enhanced.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.listener.IPoolListener;

/**
 * @author Gael Lalire
 * @param <E> pool item type
 */
public class PoolListenerDispatcherPool<E> implements IPool<E> {

    private static final Log LOG = LogFactory.getLog(PoolListenerDispatcherPool.class);

    private IPool<E> pool;

    private List<IPoolListener<E>> poolListenerList;

    public PoolListenerDispatcherPool(final IPool<E> pool) {
        this.pool = pool;
        poolListenerList = new ArrayList<IPoolListener<E>>();
    }

    public E get(final IWaitControl timeout) throws PoolException, InterruptedException {
        for (IPoolListener<E> poolListener : poolListenerList) {
            try {
                poolListener.getMethodCalled(timeout);
            } catch (RuntimeException re) {
                LOG.error("Listener failed", re);
            }
        }
        try {
            E result = pool.get(timeout);
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.getMethodReturned(result, timeout);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            return result;
        } catch (PoolException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.getMethodFailed(e, timeout);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        } catch (InterruptedException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.getMethodFailed(e, timeout);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        } catch (RuntimeException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.getMethodFailed(e, timeout);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        }
    }

    public void signalAllWaiters() {
        for (IPoolListener<E> poolListener : poolListenerList) {
            try {
                poolListener.signalAllWaitersMethodCalled();
            } catch (RuntimeException re) {
                LOG.error("Listener failed", re);
            }
        }
        try {
            pool.signalAllWaiters();
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.signalAllWaitersMethodReturned();
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
        } catch (RuntimeException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.signalAllWaitersMethodFailed(e);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        }
    }

    public void put(final E poolItem) throws NotABusyPoolItemException {
        for (IPoolListener<E> poolListener : poolListenerList) {
            try {
                poolListener.putMethodCalled(poolItem);
            } catch (RuntimeException re) {
                LOG.error("Listener failed", re);
            }
        }
        try {
            pool.put(poolItem);
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.putMethodReturned(poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
        } catch (NotABusyPoolItemException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.putMethodFailed(e, poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        } catch (RuntimeException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.putMethodFailed(e, poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        }
    }

    public void remove(final E poolItem) throws NotABusyPoolItemException {
        for (IPoolListener<E> poolListener : poolListenerList) {
            try {
                poolListener.removeMethodCalled(poolItem);
            } catch (RuntimeException re) {
                LOG.error("Listener failed", re);
            }
        }
        try {
            pool.remove(poolItem);
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.removeMethodReturned(poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
        } catch (NotABusyPoolItemException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.removeMethodFailed(e, poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        } catch (RuntimeException e) {
            for (IPoolListener<E> poolListener : poolListenerList) {
                try {
                    poolListener.removeMethodFailed(e, poolItem);
                } catch (RuntimeException re) {
                    LOG.error("Listener failed", re);
                }
            }
            throw e;
        }
    }

    public void addPoolListener(final IPoolListener<E> poolListener) {
        poolListenerList.add(poolListener);
    }

    public void removePoolListener(final IPoolListener<E> poolListener) {
        poolListenerList.remove(poolListener);
    }

}
