/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.listener.resizer;

import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.listener.IPoolListener;
import org.ow2.util.pool.impl.enhanced.api.listener.statistics.IPoolStatsListener;
import org.ow2.util.pool.impl.enhanced.internal.resizer.api.IResizer;

/**
 *
 * @author Gael Lalire
 * @param <E> pool item type
 * @param <S>
 */
public class ResizerUpdatePoolListener<E, S> implements IPoolListener<E> {

    private UpdateMethod getUpdateMethod;

    private UpdateMethod putUpdateMethod;

    private UpdateMethod removeUpdateMethod;

    private UpdateMethod interruptAllWaitersUpdateMethod;

    private IPoolStatsListener<E, S> poolStatsAccessor;

    private IResizer<S> poolResizer;

    private Object mutex;

    private S currentStats;

    public ResizerUpdatePoolListener(final S stats, final IResizer<S> poolResizer,
            final IPoolStatsListener<E, S> poolStatsAccessor) {
        this(stats, poolResizer, poolStatsAccessor, UpdateMethod.BEFORE, UpdateMethod.AFTER, UpdateMethod.AFTER,
                UpdateMethod.AFTER);
    }

    public ResizerUpdatePoolListener(final S stats, final IResizer<S> poolResizer,
            final IPoolStatsListener<E, S> poolStatsAccessor, final UpdateMethod getUpdateMethod,
            final UpdateMethod putUpdateMethod, final UpdateMethod removeUpdateMethod,
            final UpdateMethod interruptAllWaitersUpdateMethod) {
        if (stats == null) {
            throw new IllegalArgumentException();
        }
        this.poolStatsAccessor = poolStatsAccessor;
        this.poolResizer = poolResizer;
        this.currentStats = stats;
        this.getUpdateMethod = getUpdateMethod;
        this.putUpdateMethod = putUpdateMethod;
        this.removeUpdateMethod = removeUpdateMethod;
        this.interruptAllWaitersUpdateMethod = interruptAllWaitersUpdateMethod;
        mutex = new Object();
    }

    public void prepareUpdate() {
        synchronized (mutex) {
            // synchronized because latest call to copyStats must be use for preparingUpdate
            poolStatsAccessor.copyStats(currentStats);
            poolResizer.prepareUpdate(currentStats);
        }
    }

    public void update() {
        poolResizer.update();
    }

    public void getMethodCalled(final IWaitControl timeout) {
        prepareUpdate();
        if (getUpdateMethod.isBefore()) {
            update();
        }
    }

    public void getMethodFailed(final Exception poolException, final IWaitControl timeout) {
        prepareUpdate();
        if (getUpdateMethod.isAfterFailed()) {
            update();
        }
    }

    public void getMethodReturned(final E result, final IWaitControl timeout) {
        prepareUpdate();
        if (getUpdateMethod.isAfterSuccessful()) {
            update();
        }
    }

    public void signalAllWaitersMethodCalled() {
        prepareUpdate();
        if (interruptAllWaitersUpdateMethod.isBefore()) {
            update();
        }
    }

    public void putMethodCalled(final E poolItem) {
        prepareUpdate();
        if (putUpdateMethod.isBefore()) {
            update();
        }
    }

    public void putMethodFailed(final Exception poolException, final E poolItem) {
        prepareUpdate();
        if (putUpdateMethod.isAfterFailed()) {
            update();
        }
    }

    public void putMethodReturned(final E poolItem) {
        prepareUpdate();
        if (putUpdateMethod.isAfterSuccessful()) {
            update();
        }
    }

    public void removeMethodCalled(final E poolItem) {
        prepareUpdate();
        if (removeUpdateMethod.isBefore()) {
            update();
        }
    }

    public void removeMethodFailed(final Exception poolException, final E poolItem) {
        prepareUpdate();
        if (removeUpdateMethod.isAfterFailed()) {
            update();
        }
    }

    public void removeMethodReturned(final E poolItem) {
        prepareUpdate();
        if (removeUpdateMethod.isAfterSuccessful()) {
            update();
        }
    }

    public void signalAllWaitersMethodFailed(final Exception e) {
        prepareUpdate();
        if (interruptAllWaitersUpdateMethod.isAfterFailed()) {
            update();
        }
    }

    public void signalAllWaitersMethodReturned() {
        prepareUpdate();
        if (interruptAllWaitersUpdateMethod.isAfterSuccessful()) {
            update();
        }
    }

}
