/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.impl.thread;

import java.util.concurrent.Executor;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.thread.IReusableThread;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.ExpireWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.InfiniteWaitControl;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.NoWaitControl;

/**
 * @author Gael Lalire
 */
public class ThreadPool implements Executor {

    private static final Log LOG = LogFactory.getLog(ThreadPool.class);

    private IPool<IReusableThread> pool;

    private Long waitTimeBeforeUsingSelfThread;

    public ThreadPool(final IPool<IReusableThread> pool) {
        this(pool, null);
    }

    /**
     * @param pool
     * @param waitTimeBeforeUsingSelfThread if null then the threadPool will
     *        never use self thread
     */
    public ThreadPool(final IPool<IReusableThread> pool, final Long waitTimeBeforeUsingSelfThread) {
        this.pool = pool;
        this.waitTimeBeforeUsingSelfThread = waitTimeBeforeUsingSelfThread;
    }

    public void execute(final Runnable runnable) {
        boolean getCalled = false;
        IWaitControl waitControl;
        if (waitTimeBeforeUsingSelfThread == null) {
            waitControl = new InfiniteWaitControl();
        } else {
            long longValue = waitTimeBeforeUsingSelfThread.longValue();
            if (longValue <= 0) {
                waitControl = NoWaitControl.INSTANCE;
            } else {
                waitControl = new ExpireWaitControl(System.currentTimeMillis() + longValue);
            }
        }
        while (!getCalled) {
            try {
                pool.get(waitControl).setUsed(pool, runnable);
                getCalled = true;
            } catch (InterruptedException e) {
                LOG.warn("Interrupted", e);
            } catch (TimeoutPoolException e) {
                if (!waitControl.canContinueToWait()) {
                    runnable.run();
                    getCalled = true;
                } else {
                    LOG.warn("Unexpected exception", e);
                }
            } catch (PoolException e) {
                LOG.warn("PoolException", e);
            }
        }
    }

}
