/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.actionscheduler;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;


/**
 *
 * @author Gael Lalire
 */
public class ActionScheduler implements IActionScheduler {

    private static final Log LOG = LogFactory.getLog(ActionScheduler.class);

    private static final int ELEMENT_TO_KEEP_FOR_AVERAGE_COMPUTATION = 20;

    private LinkedList<Future<Boolean>> scheduledCallableList;

    private IAction action;

    private Executor executor;

    private Object mutex = new Object();

    private Average average = new Average(ELEMENT_TO_KEEP_FOR_AVERAGE_COMPUTATION);

    public ActionScheduler(final IAction action, final Executor executor) {
        this.executor = executor;
        this.action = action;
        scheduledCallableList = new LinkedList<Future<Boolean>>();
    }

    public void scheduleAction(final int time) {
        for (int i = 0; i < time; i++) {
            FutureTaskReallyCancellable<Boolean> futureTask = new FutureTaskReallyCancellable<Boolean>(action) {

                private long startTime;

                @Override
                public void run() {
                    startTime = System.currentTimeMillis();
                    super.run();
                }

                @Override
                protected void done() {
                    if (!isCancelled()) {
                        synchronized (mutex) {
                            scheduledCallableList.remove(this);
                        }
                        try {
                            if (get().booleanValue()) {
                                average.addElement(System.currentTimeMillis() - startTime);
                            }
                        } catch (InterruptedException e) {
                            LOG.error("No InterruptedException on done task expected", e);
                        } catch (ExecutionException e) {
                            LOG.error("Task failed", e);
                        }
                    }
                }

            };
            synchronized (mutex) {
                scheduledCallableList.addFirst(futureTask);
            }
            executor.execute(futureTask);
        }
    }

    public int cancelAction(final int time) {
        synchronized (mutex) {
            int canceledSheduledCallable = 0;
            Iterator<Future<Boolean>> iterator = scheduledCallableList.iterator();
            while (iterator.hasNext() && canceledSheduledCallable < time) {
                Future<Boolean> future = iterator.next();
                if (future.cancel(false)) {
                    canceledSheduledCallable++;
                    iterator.remove();
                }
            }
            return canceledSheduledCallable;
        }
    }

}
