/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.lock.impl;

import java.util.Date;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.LockSupport;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.internal.lock.api.ISignalClearableCondition;

/**
 *
 * @author Gael Lalire
 */
public class BasicCondition implements ISignalClearableCondition {

    private static final Log LOG = LogFactory.getLog(BasicCondition.class);

    private final ILockWithSignalClearableConditionControl lockControl;

    // unsignaled

    // private int unsignaledSize = 0;

    private Entry headUnsignaled;

    private Entry queueUnsignaled;

    // signaled

    // private int signaledSize = 0;

    private Entry headSignaled;

    private Entry queueUnclearableSignaled;

    private Entry queueSignaled;

    /* useful to debug
    private void calculateSize(final int expectedUnsignaledSize, final int expectedSignaledSize) {
        Entry current = headUnsignaled;
        int i = 0;
        if (current == null) {
            if (queueUnsignaled != null) {
                LOG.error("headUnsignaled==null & queueUnsignaled!=null");
            }
        } else {
            i++;
            while (current.getNext() != null) {
                i++;
                current = current.getNext();
            }
            if (current != queueUnsignaled) {
                LOG.error("queueUnsignaled!=dernier");
            }
        }
        unsignaledSize = i;
        if (i != expectedUnsignaledSize) {
            LOG.error("expectedUnsignaledSize pas bon");
        }
        current = headSignaled;
        i = 0;
        if (current == null) {
            if (queueSignaled != null) {
                LOG.error("headSignaled==null & queueSignaled!=null");
            }
        } else {
            i++;
            while (current.getNext() != null) {
                i++;
                current = current.getNext();
            }
            if (current != queueSignaled) {
                LOG.error("queueSignaled!=dernier");
            }
        }
        signaledSize = i;
        if (expectedSignaledSize != -1 && i != expectedSignaledSize) {
            LOG.error("expectedUnsignaledSize pas bon");
        }
    }
    */

    /**
     * Only one waiting thread which is unpark.
     */
    private volatile Thread effectivelySignaledThread;

    public BasicCondition(final ILockWithSignalClearableConditionControl lockControl) {
        this.lockControl = lockControl;
    }

    /**
     * @throws InterruptedException
     */
    public boolean await(final Long untilLong) throws InterruptedException {
        boolean notTimeout = true;
        lockControl.checkOwner();
        long until = 0;
        if (untilLong != null) {
            until = untilLong.longValue();
        }
        Thread current = Thread.currentThread();

        Entry next = new Entry(current);
        if (queueUnsignaled == null) {
            headUnsignaled = next;
        } else {
            queueUnsignaled.setNext(next);
        }
        queueUnsignaled = next;

        // calculateSize(unsignaledSize + 1, signaledSize);

        lockControl.innerUnlock();

        // !!! Not owner of lock

        try {
            boolean ended = false;
            do {
                if (Thread.interrupted()) {
                    throw new InterruptedException();
                }
                if (untilLong == null) {
                    LockSupport.park();
                    if (effectivelySignaledThread == current) {
                        ended = true;
                    }
                } else {
                    LockSupport.parkUntil(until);
                    if (effectivelySignaledThread == current) {
                        ended = true;
                    } else if (System.currentTimeMillis() > until) {
                        notTimeout = false;
                        ended = true;
                    }
                }
            } while (!ended);
        } finally {
            // even on InterruptedException we have to get back the lock

            lockControl.innerLock(null);

           // LOG.info("reveil");

            if (effectivelySignaledThread == current) {
                // notified (or interrupted)
                effectivelySignaledThread = null;
            } else {
                // interrupted
                next.setInterrupted(true);
            }

        }

        // Owner of lock
        return notTimeout;
    }

    public void signalAll() {
        lockControl.checkOwner();
        if (headUnsignaled != null) {
            if (queueSignaled == null) {
                headSignaled = headUnsignaled;
            } else {
                queueSignaled.setNext(headUnsignaled);
            }
            queueSignaled = queueUnsignaled;
            headUnsignaled = null;
            queueUnsignaled = null;
            lockControl.askForUnparkOneThread(signalClearableConditionControl);
            // calculateSize(0, signaledSize + unsignaledSize);
        }
    }

    public void signalAllUnclearable() {
        lockControl.checkOwner();
        if (headUnsignaled != null) {
            if (queueSignaled == null) {
                headSignaled = headUnsignaled;
            } else {
                queueSignaled.setNext(headUnsignaled);
            }
            queueSignaled = queueUnsignaled;
            headUnsignaled = null;
            queueUnsignaled = null;
            lockControl.askForUnparkOneThread(signalClearableConditionControl);
        }
        queueUnclearableSignaled = queueSignaled;
        // calculateSize(0, signaledSize + unsignaledSize);
    }

    public void signal() {
        lockControl.checkOwner();
        if (headUnsignaled != null) {
            Entry signaledEntry = headUnsignaled;
            if (queueSignaled == null) {
                headSignaled = signaledEntry;
            } else {
                queueSignaled.setNext(signaledEntry);
            }
            queueSignaled = signaledEntry;
            headUnsignaled = signaledEntry.getNext();
            if (headUnsignaled == null) {
                queueUnsignaled = null;
            }
            signaledEntry.setNext(null);
            lockControl.askForUnparkOneThread(signalClearableConditionControl);
            /*
            if (unsignaledSize == 0) {
                calculateSize(0, signaledSize);
            } else {
                calculateSize(unsignaledSize - 1, signaledSize + 1);
            }
            */
        }
    }

    public void clearAllSignal() {
        lockControl.checkOwner();
        LOG.debug("clearAllSignal");

        if (headSignaled != null) {
            Entry unsignaledEntry;
            if (queueUnclearableSignaled != null) {
                unsignaledEntry = queueUnclearableSignaled.getNext();
                queueUnclearableSignaled.setNext(null);
            } else {
                unsignaledEntry = headSignaled;
                headSignaled = null;
                lockControl.avoidAskForUnparkOneThread(signalClearableConditionControl);
            }
            Entry oldQueueSignaled = queueSignaled;
            queueSignaled = queueUnclearableSignaled;
            if (unsignaledEntry != null) {
                oldQueueSignaled.setNext(headUnsignaled);
                headUnsignaled = unsignaledEntry;
                queueUnclearableSignaled = null;
            }
        }
    }

    public void await() throws InterruptedException {
        await(null);
    }

    public boolean await(final long time, final TimeUnit unit) throws InterruptedException {
        long millisTimeout = unit.toMillis(time);
        if (millisTimeout <= 0) {
            return false;
        }
        return await(Long.valueOf(System.currentTimeMillis() + millisTimeout));
    }

    public long awaitNanos(final long nanosTimeout) throws InterruptedException {
        long lastTime = System.nanoTime();
        long millisTimeout = TimeUnit.MILLISECONDS.convert(nanosTimeout, TimeUnit.NANOSECONDS) + 1;
        await(Long.valueOf(System.currentTimeMillis() + millisTimeout));
        return nanosTimeout - (System.nanoTime() - lastTime);
    }

    public void awaitUninterruptibly() {
        boolean success = true;
        do {
            try {
                await(null);
            } catch (InterruptedException e) {
                success = false;
            }
        } while(!success);
    }

    public boolean awaitUntil(final Date deadline) throws InterruptedException {
        return await(deadline.getTime());
    }

    private ISignalClearableConditionControl signalClearableConditionControl = new ISignalClearableConditionControl() {

        public UnparkResult unparkOneThread() {
            // this method is called just before releasing the lock
            if (effectivelySignaledThread != null) {
                return UnparkResult.ALREADY_SIGNALED;
            }
            Entry unparked;
            do {
                unparked = headSignaled;
                if (unparked == null) {
                    return UnparkResult.NOTHING_TO_SIGNAL;
                }
                if (unparked == queueUnclearableSignaled) {
                    queueUnclearableSignaled = null;
                }
                headSignaled = unparked.getNext();
                if (headSignaled == null) {
                    queueSignaled = null;
                    queueUnclearableSignaled = null;
                }
            } while (unparked.isInterrupted());
            // calculateSize(unsignaledSize, -1);
          //  LOG.info("signale");
            effectivelySignaledThread = unparked.getThread();
            LockSupport.unpark(effectivelySignaledThread);
            if (headSignaled == null) {
                return UnparkResult.SIGNALED_NO_MORE_THEAD;
            }
            return UnparkResult.SIGNALED;
        }

    };


}
