/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.stoppableexecution;

import java.util.Date;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;

import org.ow2.util.pool.impl.enhanced.impl.util.LockFactory;

/**
 * @author Gael Lalire
 * @param <E> return type
 */
public class StoppableExecution<E> implements IArgumentsCallable<E> {

    private ThreadEntry head = null;

    private ThreadEntry queue = null;

    private IArgumentsCallable<E> callable;

    private Lock lock;

    private Condition condition;

    private boolean stopForced;

    public StoppableExecution(final IArgumentsCallable<E> callable) {
        lock = LockFactory.createLock();
        condition = lock.newCondition();
        this.callable = callable;
        stopForced = false;
    }

    public E call(final Object... args) throws StoppedExecutionException, Exception {
        lock.lock();
        try {
            if (stopForced) {
                throw new StoppedExecutionException();
            }
            ThreadEntry previousQueue = queue;
            ThreadEntry entry = new ThreadEntry(Thread.currentThread(), null);
            if (previousQueue != null) {
                previousQueue.setNext(entry);
            } else {
                head = entry;
            }
            queue = entry;
            try {
                lock.unlock();
                try {
                    return callable.call(args);
                } finally {
                    lock.lock();
                    // no thread death after
                }
            } catch (Error e) {
                throw new StoppedExecutionException(e);
            } finally {
                if (previousQueue == null) {
                    head = head.getNext();
                    if (head == null) {
                        condition.signalAll();
                        queue = null;
                    }
                } else {
                    if (head == entry) {
                        head = head.getNext();
                        if (head == null) {
                            condition.signalAll();
                            queue = null;
                        }
                    } else {
                        previousQueue.setNext(entry.getNext());
                    }
                }
            }
        } finally {
            lock.unlock();
        }
    }

    @SuppressWarnings("deprecation")
    public boolean forceStop(final Long expireTime) throws InterruptedException {
        Date deadline = null;
        if (expireTime != null) {
            deadline = new Date(expireTime.longValue());
        }
        lock.lock();
        try {
            stopForced = true;
            ThreadEntry current = head;
            while (current != null) {
                current.getThread().stop();
                current = current.getNext();
            }
            while (head != null) {
                if (deadline != null) {
                    if (!condition.awaitUntil(deadline)) {
                        return false;
                    }
                } else {
                    condition.await();
                }
            }
            return true;
        } finally {
            lock.unlock();
        }
    }

}
