/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.thread.impl;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.concurrent.atomic.AtomicLong;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Keep a reference on cleaner.
 * If cleaner does not have anything to clean then we keep only
 * a weak reference on cleaner. So if cleaner is not used anywhere else,
 * the cleaner is garbage collected. It produce the end of the thread running
 * this object. And normally this object should be also garbage collected.
 *
 * Warning : if you do not call add after creating a weak reference which use
 * queue of cleaner, then the thread running keeper could be terminated whereas
 * cleaner is not garbage collected. The cleaner will not continue to work but no
 * error will occurs.
 *
 * This class is final because its constructor starts a thread on its instances.
 * @author Gael Lalire
 */
public final class CleanerReferencer extends WeakReference<Cleaner> implements Runnable {

    private static final Log LOG = LogFactory.getLog(CleanerReferencer.class);

    private static AtomicLong instanceCount = new AtomicLong(0);

    private int instanceToClean;

    @SuppressWarnings("unused")
    private Cleaner cleaner;

    private Thread thread;

    private ReferenceQueue<Object> referenceQueue;

    private Object mutex;

    public CleanerReferencer(final Cleaner cleaner) {
        super(cleaner, cleaner.getReferenceQueue());
        mutex = new Object();
        this.referenceQueue = cleaner.getReferenceQueue();
        thread = new Thread(this, "Cleaner thread " + instanceCount.getAndIncrement());
        thread.setDaemon(true);
        instanceToClean = 0;
        thread.start();
    }

    public void run() {
        while (true) {
            try {
                Reference<?> reference = referenceQueue.remove();
                if (reference instanceof ICleanable) {
                    ICleanable cleanable = (ICleanable) reference;
                    try {
                        cleanable.clean();
                    } catch (RuntimeException e) {
                        e.printStackTrace();
                    }
                }
                if (remove()) {
                    break;
                }
            } catch (InterruptedException e) {
                LOG.debug("InterruptedException occurs", e);
            }
        }
    }

    public void add() {
        synchronized (mutex) {
            instanceToClean++;
            cleaner = get();
        }
    }

    private boolean remove() {
        synchronized (mutex) {
            if (instanceToClean < 0) {
                throw new IllegalStateException();
            }
            if (instanceToClean == 0) {
                return true;
            }
            instanceToClean--;
            if (instanceToClean == 0) {
                cleaner = null;
            }
            return false;
        }
    }

}
