/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.thread.impl;

import java.util.concurrent.atomic.AtomicLong;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.internal.thread.api.IManagementThread;

/**
 * Runnable can be restarted. In addition the runnable object can be garbage
 * collected. When the runnable is garbage collected this thread terminates.
 *
 * This class is final because its constructor starts a thread on its instances.
 *
 * @author Gael Lalire
 */
public final class ManagementThread extends CleanableWeakReference<Runnable> implements Runnable, IManagementThread {

    private static final Log LOG = LogFactory.getLog(ManagementThread.class);

    private static AtomicLong instanceCount = new AtomicLong(0);

    /**
     * When the runnable will be garbage collected, the
     * cleaner will call clean method.
     */
    private static Cleaner cleaner;

    static {
        cleaner = new Cleaner();
    }

    private boolean updateNeeded;

    private Object mutex;

    public ManagementThread(final Runnable runnable) {
        this(runnable, true);
    }

    public ManagementThread(final Runnable runnable, final boolean daemon) {
        super(runnable, cleaner);
        updateNeeded = false;
        mutex = new Object();
        Thread thread = new Thread(this, "Management thread " + instanceCount.getAndIncrement());
        thread.setDaemon(daemon);
        thread.start();
    }


    public void run() {
        mainloop: while (true) {
            Runnable runnable = get();
            if (runnable == null) {
                break mainloop;
            }
            synchronized (mutex) {
                while (!updateNeeded) {
                    runnable = null;
                    try {
                        mutex.wait();
                    } catch (InterruptedException e) {
                        LOG.debug("InterruptedException occurs", e);
                    }
                    runnable = get();
                    if (runnable == null) {
                        break mainloop;
                    }
                }
                updateNeeded = false;
            }
            // if start here then update will be called after this time
            // but the only way to avoid it is to keep mutex.
            // This is a bad solution because start method will be
            // slower.
            runnable.run();
        }
    }

    public void start() {
        synchronized (mutex) {
            updateNeeded = true;
            mutex.notify();
        }
    }

    public void clean() {
        synchronized (mutex) {
            mutex.notify();
        }
    }

}
