/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.pool.impl.enhanced.internal.actionscheduler;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * @author Gael Lalire
 * @param <E> return type
 */
public class FutureTaskReallyCancellable<E> implements Runnable, Future<E> {

    private Object mutex = new Object();

    private boolean cancelled;

    private boolean runned;

    private boolean runEnded;

    private Callable<E> callable;

    private E result;

    private ExecutionException exception;

    public FutureTaskReallyCancellable(final Callable<E> callable) {
        this.callable = callable;
    }

    public boolean cancel(final boolean paramBoolean) {
        synchronized (mutex) {
            if (cancelled || runned) {
                return false;
            }
            cancelled = true;
            mutex.notifyAll();
        }
        done();
        return true;
    }

    public void run() {
        synchronized (mutex) {
            if (cancelled || runned) {
                return;
            }
            runned = true;
        }
        try {
            E result = callable.call();
            synchronized (mutex) {
                runEnded = true;
                this.result = result;
                mutex.notifyAll();
            }
            done();
        } catch (Exception e) {
            synchronized (mutex) {
                runEnded = true;
                exception = new ExecutionException(e);
                mutex.notifyAll();
            }
            done();
        }
    }

    protected void done() {

    }

    public E get() throws InterruptedException, ExecutionException {
        synchronized (mutex) {
            while (!cancelled && !runEnded) {
                mutex.wait();
            }
        }
        if (exception != null) {
            throw exception;
        }
        return result;
    }

    public E get(final long paramLong, final TimeUnit paramTimeUnit) throws InterruptedException, ExecutionException,
            TimeoutException {
        long timeToWait = paramTimeUnit.toMillis(paramLong);
        if (timeToWait < 0) {
            throw new IllegalArgumentException();
        }
        if (timeToWait == 0) {
            synchronized (mutex) {
                if (!cancelled && !runEnded) {
                    throw new TimeoutException();
                }
            }
        } else {
            synchronized (mutex) {
                if (!cancelled && !runEnded) {
                    mutex.wait(timeToWait);
                }
                if (!cancelled && !runEnded) {
                    throw new TimeoutException();
                }
            }
        }
        if (exception != null) {
            throw exception;
        }
        return result;
    }

    public boolean isCancelled() {
        synchronized (mutex) {
            return cancelled;
        }
    }

    public boolean isDone() {
        synchronized (mutex) {
            if (!cancelled && !runEnded) {
                return false;
            }
        }
        return true;
    }

}
